@@ -2,9 +2,13 @@ from __future__ import (absolute_import, division, print_function,
                         unicode_literals)
 
 import collections
+from enum import Enum
+import traceback
+
 from datetime import datetime, timedelta, time as dtime
 from dateutil.parser import parse as date_parse
 import time as _time
+import trading_calendars
 import threading
 import asyncio
 
@@ -15,10 +19,11 @@ import pandas as pd
 
 import backtrader as bt
 from alpaca_trade_api.entity import Aggs
-from alpaca_trade_api.polygon.entity import NY
 from backtrader.metabase import MetaParams
 from backtrader.utils.py3 import queue, with_metaclass
 
+NY = 'America/New_York'
+
 
 # Extend the exceptions to support extra cases
 class AlpacaError(Exception):
@@ -85,6 +90,11 @@ class API(tradeapi.REST):
         return None
 
 
+class Granularity(Enum):
+    Daily = "day"
+    Minute = "minute"
+
+
 class Streamer:
     conn = None
 
@@ -96,6 +106,7 @@ class Streamer:
             instrument='',
             method='',
             base_url='',
+            data_url='',
             data_stream='',
             *args,
             **kwargs):
@@ -295,12 +306,11 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
         # poslist = positions.get('positions', [])
         return positions
 
-    def get_granularity(self, timeframe, compression):
+    def get_granularity(self, timeframe, compression) -> Granularity:
         if timeframe == bt.TimeFrame.Minutes:
-            return "minute"
+            return Granularity.Minute
         elif timeframe == bt.TimeFrame.Days:
-            return "day"
-        return None
+            return Granularity.Daily
 
     def get_instrument(self, dataname):
         try:
@@ -376,10 +386,12 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
 
     def _t_candles(self, dataname, dtbegin, dtend, timeframe, compression,
                    candleFormat, includeFirst, q):
+        granularity: Granularity = self.get_granularity(timeframe, compression)
+        dtbegin, dtend = self._make_sure_dates_are_initialized_properly(
+            dtbegin, dtend, granularity)
 
-        granularity = self.get_granularity(timeframe, compression)
         if granularity is None:
-            e = AlpacaTimeFrameError('')
+            e = AlpacaTimeFrameError('granularity is missing')
             q.put(e.error_response)
             return
         try:
@@ -387,21 +399,21 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
                 cdl = self.get_aggs_from_polygon(dataname,
                                                  dtbegin,
                                                  dtend,
-                                                 granularity,
+                                                 granularity.value,
                                                  compression)
             else:
                 cdl = self.get_aggs_from_alpaca(dataname,
                                                 dtbegin,
                                                 dtend,
-                                                granularity,
+                                                granularity.value,
                                                 compression)
         except AlpacaError as e:
             print(str(e))
             q.put(e.error_response)
             q.put(None)
             return
-        except Exception as e:
-            print(str(e))
+        except Exception:
+            traceback.print_exc()
             q.put({'code': 'error'})
             q.put(None)
             return
@@ -409,8 +421,10 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
         # don't use dt.replace. use localize
         # (https://stackoverflow.com/a/1592837/2739124)
         cdl = cdl.loc[
-              pytz.timezone(NY).localize(dtbegin):
-              pytz.timezone(NY).localize(dtend)
+              pytz.timezone(NY).localize(dtbegin) if
+              not dtbegin.tzname() else dtbegin:
+              pytz.timezone(NY).localize(dtend) if
+              not dtend.tzname() else dtend
               ].dropna(subset=['high'])
         records = cdl.reset_index().to_dict('records')
         for r in records:
@@ -418,6 +432,49 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
             q.put(r)
         q.put({})  # end of transmission
 
+    def _make_sure_dates_are_initialized_properly(self, dtbegin, dtend,
+                                                  granularity):
+        """
+        dates may or may not be specified by the user.
+        when they do, they are probably don't include NY timezome data
+        also, when granularity is minute, we want to make sure we get data when
+        market is opened. so if it doesn't - let's get set end date to be last
+        known minute with opened market.
+        this nethod takes care of all these issues.
+        :param dtbegin:
+        :param dtend:
+        :param granularity:
+        :return:
+        """
+        if not dtend:
+            dtend = pd.Timestamp('now', tz=NY)
+        else:
+            dtend = pd.Timestamp(pytz.timezone(NY).localize(dtend))
+        if granularity == Granularity.Minute:
+            calendar = trading_calendars.get_calendar(name='NYSE')
+            if pd.Timestamp('now', tz=NY).date() == dtend.date():
+                if calendar.is_open_on_minute(dtend):
+                    # we execute during market open, we don't want today's data
+                    # we will receive it through the websocket
+                    dtend = dtend.replace(hour=15,
+                                          minute=59,
+                                          second=0,
+                                          microsecond=0)
+                    dtend -= timedelta(days=1)
+            while not calendar.is_open_on_minute(dtend):
+                dtend = dtend.replace(hour=15,
+                                      minute=59,
+                                      second=0,
+                                      microsecond=0)
+                dtend -= timedelta(days=1)
+        if not dtbegin:
+            days = 30 if granularity == Granularity.Daily else 3
+            delta = timedelta(days=days)
+            dtbegin = dtend - delta
+        else:
+            dtbegin = pd.Timestamp(pytz.timezone(NY).localize(dtbegin))
+        return dtbegin, dtend
+
     def get_aggs_from_polygon(self,
                               dataname,
                               dtbegin,
@@ -443,13 +500,6 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
             """
             return df.between_time("09:30", "16:00")
 
-        if not dtend:
-            dtend = datetime.utcnow()
-        if not dtbegin:
-            days = 30 if 'd' in granularity else 3
-            delta = timedelta(days=days)
-            dtbegin = dtend - delta
-
         if granularity == 'day':
             cdl = self.oapi.polygon.historic_agg_v2(
                 dataname,
@@ -462,7 +512,7 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
             segment_start = dtbegin
             segment_end = segment_start + timedelta(weeks=2) if \
                 dtend - dtbegin >= timedelta(weeks=2) else dtend
-            while segment_end <= dtend:
+            while segment_end <= dtend and dtend not in cdl.index:
                 response = self.oapi.polygon.historic_agg_v2(
                     dataname,
                     compression,
@@ -470,7 +520,7 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
                     _from=self.iso_date(segment_start.isoformat()),
                     to=self.iso_date(segment_end.isoformat()))
                 # No result from the server, most likely error
-                if response.df.shape[0] == 0:
+                if response.df.shape[0] == 0 and cdl.shape[0] == 0:
                     raise Exception("received empty response")
                 temp = response.df
                 cdl = pd.concat([cdl, temp])
@@ -532,16 +582,20 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
                                          limit=1000,
                                          end=curr.isoformat()
                                          )[dataname]
-                earliest_sample = r[0].t
-                r = r._raw
-                r.extend(response)
-                response = r
-                if earliest_sample <= pytz.timezone(NY).localize(start):
-                    got_all = True
+                if r:
+                    earliest_sample = r[0].t
+                    r = r._raw
+                    r.extend(response)
+                    response = r
+                    if earliest_sample <= pytz.timezone(NY).localize(start):
+                        got_all = True
+                    else:
+                        delta = timedelta(days=1) if granularity == "day" \
+                            else timedelta(minutes=1)
+                        curr = earliest_sample - delta
                 else:
-                    delta = timedelta(days=1) if granularity == "day" else \
-                        timedelta(minutes=1)
-                    curr = earliest_sample - delta
+                    # no more data is available, let's return what we have
+                    break
             return response
 
         def _clear_out_of_market_hours(df):
@@ -655,6 +709,7 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
         bt.Order.Limit: 'limit',
         bt.Order.Stop: 'stop',
         bt.Order.StopLimit: 'stop_limit',
+        bt.Order.StopTrail: 'trailing_stop',
     }
 
     def broker_threads(self):
@@ -712,7 +767,7 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
         okwargs['side'] = 'buy' if order.isbuy() else 'sell'
         okwargs['type'] = self._ORDEREXECS[order.exectype]
         okwargs['time_in_force'] = "gtc"
-        if order.exectype != bt.Order.Market:
+        if order.exectype not in [bt.Order.Market, bt.Order.StopTrail]:
             okwargs['limit_price'] = str(order.created.price)
 
         if order.exectype in [bt.Order.StopLimit, bt.Order.Stop]:
@@ -731,6 +786,18 @@ class AlpacaStore(with_metaclass(MetaSingleton, object)):
         if stopside or takeside:
             okwargs['order_class'] = "bracket"
 
+        if order.exectype == bt.Order.StopTrail:
+            if order.trailpercent and order.trailamount:
+                raise Exception("You can't create trailing stop order with "
+                                "both TrailPrice and TrailPercent. choose one")
+            if order.trailpercent:
+                okwargs['trail_percent'] = order.trailpercent
+            elif order.trailamount:
+                okwargs['trail_price'] = order.trailamount
+            else:
+                raise Exception("You must provide either trailpercent or "
+                                "trailamount when creating StopTrail order")
+
         okwargs.update(**kwargs)  # anything from the user
 
         self.q_ordercreate.put((order.ref, okwargs,))
